/*
 * LevelDB.cpp
 *
 *  Created on: Jul 10, 2014
 *      Author: lion
 */

#include "object.h"
#include "LevelDB.h"
#include "ifs/db.h"
#include "Buffer.h"

namespace fibjs {

result_t db_base::openLevelDB(exlib::string connString,
    obj_ptr<LevelDB_base>& retVal, AsyncEvent* ac)
{
    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    result_t hr;
    const char* c_str = connString.c_str();

    if (!qstrcmp(c_str, "leveldb:", 8))
        c_str += 8;

    obj_ptr<LevelDB> db = new LevelDB();
    hr = db->open(c_str);
    if (hr < 0)
        return hr;

    retVal = db;

    return 0;
}

result_t LevelDB::open(const char* connString)
{
    leveldb::Options options;
    options.create_if_missing = true;

#ifdef _WIN32
    exlib::wstring wstr = utf8to16String(connString);
    exlib::string str;

    int chars_num = (int)WideCharToMultiByte(CP_ACP, 0, (const wchar_t*)wstr.c_str(), (int)wstr.length(), NULL, 0, NULL, NULL);
    str.resize(chars_num);
    WideCharToMultiByte(CP_ACP, 0, (const wchar_t*)wstr.c_str(), (int)wstr.length(), str.data(), chars_num + 1, NULL, NULL);
    connString = str.c_str();
#endif

    leveldb::Status s = leveldb::DB::Open(options, connString, &m_db);
    if (!s.ok())
        return CHECK_ERROR(error(s));

    return 0;
}

result_t close_ldb(leveldb::DB* db)
{
    delete db;
    return 0;
}

LevelDB::~LevelDB()
{
    if (m_batch) {
        m_batch->Clear();
        delete m_batch;
    } else if (m_db)
        asyncCall(close_ldb, m_db);
}

result_t LevelDB::has(Buffer_base* key, bool& retVal, AsyncEvent* ac)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    Buffer* buf = (Buffer*)key;

    std::string value;
    leveldb::Status s = db()->Get(leveldb::ReadOptions(),
        leveldb::Slice((const char*)buf->data(), buf->length()),
        &value);
    if (s.IsNotFound()) {
        retVal = false;
        return 0;
    } else if (!s.ok())
        return CHECK_ERROR(error(s));

    retVal = true;

    return 0;
}

result_t LevelDB::get(Buffer_base* key, obj_ptr<Buffer_base>& retVal, AsyncEvent* ac)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    Buffer* buf = (Buffer*)key;

    std::string value;
    leveldb::Status s = db()->Get(leveldb::ReadOptions(),
        leveldb::Slice((const char*)buf->data(), buf->length()),
        &value);
    if (s.IsNotFound())
        return CALL_RETURN_NULL;
    else if (!s.ok())
        return CHECK_ERROR(error(s));

    retVal = new Buffer(value.c_str(), value.length());

    return 0;
}

result_t LevelDB::_mget(std::vector<exlib::string>* keys,
    obj_ptr<NArray>& retVal, AsyncEvent* ac)
{
    std::vector<exlib::string>& ks = *keys;
    obj_ptr<NArray> list = new NArray();
    int32_t i;

    Variant nil;
    nil.setNull();

    for (i = 0; i < (int32_t)ks.size(); i++) {
        std::string value;

        leveldb::Status s = db()->Get(leveldb::ReadOptions(),
            leveldb::Slice(ks[i].c_str(), ks[i].length()),
            &value);
        if (s.IsNotFound())
            list->append(nil);
        else if (!s.ok())
            return CHECK_ERROR(error(s));
        else {
            obj_ptr<Buffer_base> buf = new Buffer(value.c_str(), value.length());
            list->append(buf);
        }
    }

    retVal = list;

    return 0;
}

result_t LevelDB::mget(v8::Local<v8::Array> keys, obj_ptr<NArray>& retVal)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    std::vector<exlib::string> ks;
    int32_t len = keys->Length();
    int32_t i;
    result_t hr;

    if (!len)
        return CALL_RETURN_NULL;

    ks.resize(len);

    Isolate* isolate = holder();
    v8::Local<v8::Context> context = isolate->context();

    for (i = 0; i < len; i++) {
        JSValue v = keys->Get(context, i);
        obj_ptr<Buffer_base> buf;

        hr = GetArgumentValue(isolate, v, buf);
        if (hr < 0)
            return CHECK_ERROR(hr);

        exlib::string s;
        buf->toString(s);

        ks[i] = s;
    }

    return ac__mget(&ks, retVal);
}

result_t LevelDB::_commit(leveldb::WriteBatch* batch, AsyncEvent* ac)
{
    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    leveldb::Status s = db()->Write(leveldb::WriteOptions(), batch);
    if (!s.ok())
        return CHECK_ERROR(error(s));

    return 0;
}

result_t LevelDB::set(Buffer_base* key, Buffer_base* value, AsyncEvent* ac)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    Buffer* buf_key = (Buffer*)key;
    Buffer* buf_val = (Buffer*)value;

    leveldb::Status s = Set(leveldb::Slice((const char*)buf_key->data(), buf_key->length()),
        leveldb::Slice((const char*)buf_val->data(), buf_val->length()));
    if (!s.ok())
        return CHECK_ERROR(error(s));

    return 0;
}

result_t LevelDB::mset(v8::Local<v8::Object> map)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    leveldb::WriteBatch batch;
    leveldb::WriteBatch* batch_ = m_batch ? m_batch : &batch;

    Isolate* isolate = holder();
    v8::Local<v8::Context> context = isolate->context();
    JSArray ks = map->GetPropertyNames(context);
    int32_t len = ks->Length();
    int32_t i;
    result_t hr;

    for (i = 0; i < len; i++) {
        JSValue k = ks->Get(context, i);
        exlib::string key(isolate->toString(k));

        exlib::string value1;
        hr = getValue(JSValue(map->Get(context, k)), value1);
        if (hr < 0)
            return hr;

        batch_->Put(leveldb::Slice(key.c_str(), key.length()),
            leveldb::Slice(value1.c_str(), value1.length()));
    }

    if (m_batch)
        return 0;

    return ac__commit(&batch);
}

result_t LevelDB::mremove(v8::Local<v8::Array> keys)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    Isolate* isolate = holder();
    v8::Local<v8::Context> context = isolate->context();
    leveldb::WriteBatch batch;
    leveldb::WriteBatch* batch_ = m_batch ? m_batch : &batch;

    int32_t len = keys->Length();
    int32_t i;
    result_t hr;

    for (i = 0; i < len; i++) {
        exlib::string key;
        hr = getValue(JSValue(keys->Get(context, i)), key);
        if (hr < 0)
            return hr;

        batch_->Delete(leveldb::Slice(key.c_str(), key.length()));
    }

    if (m_batch)
        return 0;

    return ac__commit(&batch);
}

result_t LevelDB::remove(Buffer_base* key, AsyncEvent* ac)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    Buffer* buf = (Buffer*)key;

    exlib::string value;
    leveldb::Status s = Delete(leveldb::Slice((const char*)buf->data(), buf->length()));
    if (!s.ok())
        return CHECK_ERROR(error(s));

    return 0;
}

result_t LevelDB::firstKey(obj_ptr<Buffer_base>& retVal, AsyncEvent* ac)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    leveldb::Iterator* it = db()->NewIterator(leveldb::ReadOptions());
    it->SeekToFirst();
    if (!it->Valid()) {
        delete it;
        return CALL_RETURN_NULL;
    }

    retVal = new Buffer(it->key().data(), it->key().size());
    delete it;

    return 0;
}

result_t LevelDB::lastKey(obj_ptr<Buffer_base>& retVal, AsyncEvent* ac)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    leveldb::Iterator* it = db()->NewIterator(leveldb::ReadOptions());
    it->SeekToLast();
    if (!it->Valid()) {
        delete it;
        return CALL_RETURN_NULL;
    }

    retVal = new Buffer(it->key().data(), it->key().size());
    delete it;

    return 0;
}

result_t LevelDB::Iter::_iter(AsyncEvent* ac)
{
    m_count = 0;

    if (m_first) {
        if (m_reverse) {
            if (m_from.empty())
                m_it->SeekToLast();
            else {
                m_it->Seek(leveldb::Slice(m_from.c_str(), m_from.length()));

                leveldb::Slice key = m_it->key();
                if (key.compare(leveldb::Slice(m_from.c_str(), m_from.length())) != 0)
                    m_it->Prev();
            }
        } else {
            if (m_from.empty())
                m_it->SeekToFirst();
            else
                m_it->Seek(leveldb::Slice(m_from.c_str(), m_from.length()));
        }

        m_first = false;

        if (!m_it->Valid()) {
            m_end = true;
            return 0;
        }
    }

    while (m_count < ITER_BLOCK_SIZE) {
        leveldb::Slice key = m_it->key();
        if (!m_to.empty() && key.compare(leveldb::Slice(m_to.c_str(), m_to.length())) >= 0) {
            m_end = true;
            break;
        }

        if (m_skip == 0) {
            m_kvs[m_count * 2] = new Buffer(key.data(), key.size());
            leveldb::Slice value = m_it->value();
            m_kvs[m_count * 2 + 1] = new Buffer(value.data(), value.size());

            m_count++;

            if (m_limit > 0) {
                m_limit--;
                if (m_limit == 0) {
                    m_end = true;
                    break;
                }
            }
        } else
            m_skip--;

        if (m_reverse)
            m_it->Prev();
        else
            m_it->Next();

        if (!m_it->Valid()) {
            m_end = true;
            break;
        }
    }

    return 0;
}

result_t LevelDB::Iter::iter(Isolate* isolate, v8::Local<v8::Function> func)
{
    result_t hr;
    int32_t i;

    do {
        hr = ac__iter();
        if (hr < 0)
            return hr;

        for (i = 0; i < m_count; i++) {
            v8::EscapableHandleScope handle_scope(isolate->m_isolate);

            v8::Local<v8::Value> args[2] = {
                m_kvs[i * 2 + 1]->wrap(), m_kvs[i * 2]->wrap()
            };

            m_kvs[i * 2].Release();
            m_kvs[i * 2 + 1].Release();

            v8::Local<v8::Value> v = func->Call(func->GetCreationContextChecked(), v8::Undefined(isolate->m_isolate), 2, args).FromMaybe(v8::Local<v8::Value>());
            if (v.IsEmpty())
                return CALL_E_JAVASCRIPT;

            if (isolate->toBoolean(v)) {
                m_end = true;
                break;
            }
        }
    } while (!m_end);

    return 0;
}

result_t LevelDB::forEach(v8::Local<v8::Function> func)
{
    if (!db())
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    return Iter(db()).iter(holder(), func);
}

result_t LevelDB::forEach(Buffer_base* from, v8::Local<v8::Function> func)
{
    return forEach(from, NULL, v8::Local<v8::Object>(), func);
}

result_t LevelDB::forEach(Buffer_base* from, Buffer_base* to, v8::Local<v8::Function> func)
{
    return forEach(from, to, v8::Local<v8::Object>(), func);
}

result_t LevelDB::forEach(v8::Local<v8::Object> opt, v8::Local<v8::Function> func)
{
    return forEach(NULL, NULL, opt, func);
}

result_t LevelDB::forEach(Buffer_base* from, v8::Local<v8::Object> opt, v8::Local<v8::Function> func)
{
    return forEach(from, NULL, opt, func);
}

result_t LevelDB::forEach(Buffer_base* from, Buffer_base* to, v8::Local<v8::Object> opt, v8::Local<v8::Function> func)
{
    obj_ptr<Iter> it = new Iter(db());

    it->getValue(from, to);
    if (!opt.IsEmpty()) {
        Isolate* isolate = holder();
        result_t hr;

        hr = GetConfigValue(isolate, opt, "skip", it->m_skip, true);
        if (hr < 0 && hr != CALL_E_PARAMNOTOPTIONAL)
            return hr;

        hr = GetConfigValue(isolate, opt, "limit", it->m_limit, true);
        if (hr < 0 && hr != CALL_E_PARAMNOTOPTIONAL)
            return hr;
        if (it->m_limit == 0)
            return Runtime::setError("limit must be greater than 0");

        hr = GetConfigValue(isolate, opt, "reverse", it->m_reverse, true);
        if (hr < 0 && hr != CALL_E_PARAMNOTOPTIONAL)
            return hr;
    }

    return it->iter(holder(), func);
}

result_t LevelDB::begin(obj_ptr<LevelDB_base>& retVal)
{
    if (!m_db)
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    obj_ptr<LevelDB> db = new LevelDB();

    db->m_base = this;
    db->m_batch = new leveldb::WriteBatch();

    retVal = db;

    return 0;
}

result_t LevelDB::commit()
{
    if (!db() || !m_batch)
        return CHECK_ERROR(CALL_E_INVALID_CALL);

    result_t hr = ac__commit(m_batch);
    cc_close();

    return hr;
}

result_t LevelDB::close(AsyncEvent* ac)
{
    if (m_batch) {
        m_base.Release();

        m_batch->Clear();
        delete m_batch;
        m_batch = NULL;

        return 0;
    }

    if (!m_db)
        return 0;

    if (ac->isSync())
        return CHECK_ERROR(CALL_E_NOSYNC);

    delete m_db;
    m_db = NULL;

    return 0;
}
}
